from kurbopy import BezPath, Point
import math


def tangent(seg, t):
    return seg.deriv().eval(t)


def wonkiness(path: BezPath):
    def nxt(lst, n):
        return lst[(n + 1) % len(lst)]

    segs = path.segments()
    segscore = 0
    for i, seg in enumerate(segs):
        in_curvature = seg.curvature(0.9)
        out_curvature = nxt(segs, i).curvature(0.1)
        curvaturediff = abs(1000 * (in_curvature + out_curvature))
        # print("In curve: %e Out curve: %e" % (in_curvature, out_curvature))
        in_tangent = tangent(seg, 0.9)
        out_tangent = tangent(nxt(segs, i), 0.1)
        anglediff = abs(
            abs(in_tangent.to_vec2().atan2()) - abs(out_tangent.to_vec2().atan2())
        ) % (math.pi / 4)
        # print(
        #     "Seg %i Angle diff: %.9e Curve diff: %.9e" % (i, anglediff, curvaturediff)
        # )
        total_len = seg.arclen(0.1) + nxt(segs, i).arclen(0.1)
        segscore += 100 * (anglediff + curvaturediff) / (1 + total_len)
    print("Total score", segscore)
    return segscore / len(segs)


b = BezPath()
b.move_to(Point(0, 0))
b.line_to(Point(100, 0))
b.line_to(Point(100, 100))
b.line_to(Point(90, 100))
b.line_to(Point(90, 90))
b.line_to(Point(75, 150))
b.close_path()
import matplotlib.pyplot as plt

fig, ax = plt.subplots()
b.plot(ax)
plt.show()
print(list(f for f in b.segments()))
print(wonkiness(b))



from .kurbopy import Affine
from .kurbopy import Arc
from .kurbopy import BezPath
from .kurbopy import Circle
from .kurbopy import CircleSegment
from .kurbopy import ConstPoint
from .kurbopy import CubicBez
# CubicOffset XXX
# CurveFitSample XXX
from .kurbopy import Ellipse
from .kurbopy import Insets
from .kurbopy import Line
from .kurbopy import LineIntersection
from .kurbopy import MinDistance
from .kurbopy import Nearest
from .kurbopy import PathEl
from .kurbopy import PathSeg
from .kurbopy import Point
from .kurbopy import QuadBez
from .kurbopy import QuadSpline
from .kurbopy import Rect
# RoundedRect XXX
# RoundedRectRadii XXX
# Segments XXX
from .kurbopy import Size
# Stroke XXX
# StrokeOpts XXX
# SVGArc XXX
from .kurbopy import TranslateScale
from .kurbopy import Vec2
from .kurbopy import cubics_to_quadratic_splines, offset_cubic
from fontTools.pens.basePen import BasePen
from kurbopy.magic import magic_mul, magic_add, magic_sub
import re


def from_drawable(drawable, *penArgs, **penKwargs):
    """Returns an *array of BezPath* from any object conforming to the pen protocol."""
    pen = BezPathCreatingPen(*penArgs, **penKwargs)
    drawable.draw(pen)
    return pen.paths


setattr(BezPath, "from_drawable", from_drawable)


def to_matplot(self):
    from matplotlib.path import Path

    svg = self.to_svg()
    verts = []
    codes = []
    while len(svg):
        m = re.match(r"^M(-?[\d\.]+)[,\s](-?[\d\.]+)\s*", svg)
        if m:
            codes.append(Path.MOVETO)
            verts.append((m[1], m[2]))
            svg = svg[len(m[0]) :]
            continue

        m = re.match(r"^L(-?[\d\.]+)[,\s](-?[\d\.]+)\s*", svg)
        if m:
            codes.append(Path.LINETO)
            verts.append((m[1], m[2]))
            svg = svg[len(m[0]) :]
            continue

        m = re.match(
            r"^Q(-?[\d\.]+)[,\s](-?[\d\.]+) (-?[\d\.]+)[,\s](-?[\d\.]+)\s*", svg
        )
        if m:
            codes.append(Path.CURVE3)
            verts.append((m[1], m[2]))
            codes.append(Path.CURVE3)
            verts.append((m[3], m[4]))
            svg = svg[len(m[0]) :]
            continue

        m = re.match(
            r"^C(-?[\d\.]+)[,\s](-?[\d\.]+) (-?[\d\.]+)[,\s](-?[\d\.]+) (-?[\d\.]+)[,\s](-?[\d\.]+)\s*",
            svg,
        )
        if m:
            codes.append(Path.CURVE4)
            verts.append((m[1], m[2]))
            codes.append(Path.CURVE4)
            verts.append((m[3], m[4]))
            codes.append(Path.CURVE4)
            verts.append((m[5], m[6]))
            svg = svg[len(m[0]) :]
            continue

        if svg[0] == "Z":
            verts.append(verts[0])
            codes.append(Path.CLOSEPOLY)
            svg = svg[1:]
            continue
        raise (ValueError("Don't know what to do with " + svg))

    return Path(verts, codes)


setattr(BezPath, "to_matplot", to_matplot)


def plot(self, ax, **kwargs):
    """Plot the path on a Matplot subplot which you supply

    ::

          import matplotlib.pyplot as plt
          fig, ax = plt.subplots()
          path.plot(ax)

    """
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    from matplotlib.path import Path
    import matplotlib.patches as patches

    path = self.to_matplot()
    if not "lw" in kwargs:
        kwargs["lw"] = 2
    if not "fill" in kwargs:
        kwargs["fill"] = False
    drawNodes = not ("drawNodes" in kwargs) or kwargs["drawNodes"] != False
    if "drawNodes" in kwargs:
        kwargs.pop("drawNodes")
    patch = patches.PathPatch(path, **kwargs)
    ax.add_patch(patch)
    left, right = ax.get_xlim()
    top, bottom = ax.get_ylim()
    bounds = self.bounding_box()
    bounds = bounds.inset(Insets.uniform(50))
    if not (left == 0.0 and right == 1.0 and top == 0.0 and bottom == 1.0):
        bounds = bounds.union_pt(Point(left, top))
        bounds = bounds.union_pt(Point(right, bottom))
    ax.set_xlim(bounds.min_x(), bounds.max_x())
    ax.set_ylim(bounds.min_y(), bounds.max_y())


setattr(BezPath, "plot", plot)


class BezPathCreatingPen(BasePen):
    def __init__(self, *args, **kwargs):
        super(BezPathCreatingPen, self).__init__(*args, **kwargs)
        self.paths = []
        self.path = BezPath()

    def _moveTo(self, p):
        self.path.move_to(Point(p[0], p[1]))

    def _lineTo(self, p):
        self.path.line_to(Point(p[0], p[1]))

    def _curveToOne(self, p1, p2, p3):
        self.path.curve_to(
            Point(p1[0], p1[1]), Point(p2[0], p2[1]), Point(p3[0], p3[1])
        )

    def _qCurveToOne(self, p1, p2):
        self.path.quad_to(Point(p1[0], p1[1]), Point(p2[0], p2[1]))

    def _closePath(self):
        self.path.close_path()
        self.paths.append(self.path)
        self.path = BezPath()



from kurbopy import Rect, Point, Size, Insets


def test_positive_insets():
    rect = Rect.from_origin_size(
        Point(
            0.0,
            0.0,
        ),
        Size(
            10.0,
            10.0,
        ),
    )
    insets = Insets.uniform_xy(3.0, 0.0)
    inset_rect = rect + insets
    assert inset_rect.width() == 16.0, "10.0 + 3.0 x 2"
    assert inset_rect.x0 == -3.0


def test_negative_insets():
    rect = Rect.from_origin_size(
        Point(
            0.0,
            0.0,
        ),
        Size(
            10.0,
            10.0,
        ),
    )
    insets = Insets.uniform_xy(-3.0, 0.0)
    inset_rect = rect + insets
    assert inset_rect.width() == 4.0, "10.0 - 3.0 × 2"
    assert inset_rect.x0 == 3.0


def test_absolute():
    rect = Rect(7.0, 11.0, 0.0, 0.0)
    insets = Insets.uniform_xy(0.0, 1.0)
    assert rect.width() == -7.0
    inset_rect = rect + insets
    assert inset_rect.width() == 7.0
    assert inset_rect.x0 == 0.0
    assert inset_rect.height() == 13.0


def test_rect_rect():
    rect = Rect(0.0, 0.0, 5.0, 11.0)
    insets = Insets.uniform_xy(
        1.0,
        7.0,
    )
    inset_rect = rect + insets
    insets2 = inset_rect - rect

    assert insets2.x0 == insets.x0
    assert insets2.y1 == insets.y1
    assert insets2.x_value() == insets.x_value()
    assert insets2.y_value() == insets.y_value()



from kurbopy import Point, BezPath
import math


def test_bezpath_segments():
    b = BezPath()
    b.move_to(Point(0, 0))
    b.line_to(Point(100, 100))
    b.line_to(Point(100, 0))
    b.close_path()
    assert len(b.segments()) == 3



import math
from kurbopy import Arc, Point, Vec2


def test_arc_iterator():
    arc = Arc(Point(1, 1), Vec2(1, 0), math.pi / 2, math.pi, 0)
    lst = []
    arc.to_cubic_beziers(0.5, lambda a, b, c: lst.append((a, b, c)))
    assert len(lst) == 2



import functools
import sys


# This module allows for polymorphic operator overloading between objects of
# different classes without driving Rust's type system mad. When a magic method
# (eg. __add__) is called on a Rust object, the Rust object will call
# `kurbopy.magic.magic_add(self, rhs)`. This will in turn look up a method on
# the Rust object, `_add_Rhs`, where `Rhs` is the class name of the rhs object.
# If this method is not found, a TypeError is raised.


def get_magic_name(obj):
    cls = obj.__class__
    if hasattr(cls, "__name__"):
        return cls.__name__
    raise ValueError("No magic name for " + str(obj))


def do_magic(self, rhs, methodname):
    other_type = get_magic_name(rhs)
    if hasattr(self, "_" + methodname + "_" + other_type):
        return getattr(self, "_" + methodname + "_" + other_type)(rhs)
    mytype = get_magic_name(self)
    raise TypeError("Cannot %s %s by %s" % (methodname, mytype, other_type))


for method in ["mul", "add", "sub", "isub", "iadd"]:
    magic = functools.partial(do_magic, methodname=method)
    magic.__name__ = method
    magic.__doc__ = "Magic method " + method
    globals()["magic_" + method] = magic




from kurbopy import Point, CubicBez
import math


def test_cubicbez_deriv():
    c = CubicBez(
        Point(0.0, 0.0),
        Point(1.0 / 3.0, 0.0),
        Point(2.0 / 3.0, 1.0 / 3.0),
        Point(1.0, 1.0),
    )
    deriv = c.deriv()

    n = 10
    for i in range(1, n):
        t = 1 / (i * n)
        delta = 1e-6
        p = c.eval(t)
        p1 = c.eval(t + delta)
        d_approx = (p1.to_vec2() - p.to_vec2()) * (1 / delta)
        d = deriv.eval(t).to_vec2()
        assert (d - d_approx).hypot() < delta * 2.0


def test_cubicbez_arclen():
    # y = x^2
    c = CubicBez(
        Point(0.0, 0.0),
        Point(1.0 / 3.0, 0.0),
        Point(2.0 / 3.0, 1.0 / 3.0),
        Point(1.0, 1.0),
    )
    true_arclen = 0.5 * math.sqrt(5.0) + 0.25 * math.log(2.0 + math.sqrt(5.0))
    for i in range(0, 12):
        accuracy = 0.1**i
        error = c.arclen(accuracy) - true_arclen
        assert abs(error) < accuracy


# def test_cubicbez_inv_arclen():
#     // y = x^2 / 100
#     c = CubicBez(
#         Point(0.0, 0.0),
#         Point(100.0 / 3.0, 0.0),
#         Point(200.0 / 3.0, 100.0 / 3.0),
#         Point(100.0, 100.0),
#     );
#     true_arclen = 100.0 * (0.5 * 5.0f64.sqrt() + 0.25 * (2.0 + 5.0f64.sqrt()).ln());
#     for i in 0..12 {
#         accuracy = 0.1f64.powi(i);
#         n = 10;
#         for j in 0..=n {
#             arc = (j as f64) * ((n as f64).recip() * true_arclen);
#             t = c.inv_arclen(arc, accuracy * 0.5);
#             actual_arc = c.subsegment(0.0..t).arclen(accuracy * 0.5);
#             assert!(
#                 (arc - actual_arc).abs() < accuracy,
#                 "at accuracy {:e, wanted { got {",
#                 accuracy,
#                 actual_arc,
#                 arc
#             );


#     // corner case: user passes accuracy larger than total arc length
#     accuracy = true_arclen * 1.1;
#     arc = true_arclen * 0.5;
#     t = c.inv_arclen(arc, accuracy);
#     actual_arc = c.subsegment(0.0..t).arclen(accuracy);
#     assert!(
#         (arc - actual_arc).abs() < 2.0 * accuracy,
#         "at accuracy {:e, want { got {",
#         accuracy,
#         actual_arc,
#         arc
#     );


# def test_cubicbez_signed_area_linear():
#     #
#     c = CubicBez::new(
#         (1.0, 0.0),
#         (2.0 / 3.0, 1.0 / 3.0),
#         (1.0 / 3.0, 2.0 / 3.0),
#         (0.0, 1.0),
#     );
#     epsilon = 1e-12;
#     assert_eq!((Affine::rotate(0.5) * c).signed_area(), 0.5);
#     assert!(((Affine::rotate(0.5) * c).signed_area() - 0.5).abs() < epsilon);
#     assert!(((Affine::translate((0.0, 1.0)) * c).signed_area() - 1.0).abs() < epsilon);
#     assert!(((Affine::translate((1.0, 0.0)) * c).signed_area() - 1.0).abs() < epsilon);


# def test_cubicbez_signed_area():
#     // y = 1 - x^3
#     c = CubicBez::new((1.0, 0.0), (2.0 / 3.0, 1.0), (1.0 / 3.0, 1.0), (0.0, 1.0));
#     epsilon = 1e-12;
#     assert!((c.signed_area() - 0.75).abs() < epsilon);
#     assert!(((Affine::rotate(0.5) * c).signed_area() - 0.75).abs() < epsilon);
#     assert!(((Affine::translate((0.0, 1.0)) * c).signed_area() - 1.25).abs() < epsilon);
#     assert!(((Affine::translate((1.0, 0.0)) * c).signed_area() - 1.25).abs() < epsilon);


# def test_cubicbez_nearest():
#     fn verify(result: Nearest, expected: f64) {
#         assert!(
#             (result.t - expected).abs() < 1e-6,
#             "got {:? expected {",
#             result,
#             expected
#         );

#     // y = x^3
#     c = CubicBez::new((0.0, 0.0), (1.0 / 3.0, 0.0), (2.0 / 3.0, 0.0), (1.0, 1.0));
#     verify(c.nearest((0.1, 0.001).into(), 1e-6), 0.1);
#     verify(c.nearest((0.2, 0.008).into(), 1e-6), 0.2);
#     verify(c.nearest((0.3, 0.027).into(), 1e-6), 0.3);
#     verify(c.nearest((0.4, 0.064).into(), 1e-6), 0.4);
#     verify(c.nearest((0.5, 0.125).into(), 1e-6), 0.5);
#     verify(c.nearest((0.6, 0.216).into(), 1e-6), 0.6);
#     verify(c.nearest((0.7, 0.343).into(), 1e-6), 0.7);
#     verify(c.nearest((0.8, 0.512).into(), 1e-6), 0.8);
#     verify(c.nearest((0.9, 0.729).into(), 1e-6), 0.9);
#     verify(c.nearest((1.0, 1.0).into(), 1e-6), 1.0);
#     verify(c.nearest((1.1, 1.1).into(), 1e-6), 1.0);
#     verify(c.nearest((-0.1, 0.0).into(), 1e-6), 0.0);
#     a = Affine::rotate(0.5);
#     verify((a * c).nearest(a * Point::new(0.1, 0.001), 1e-6), 0.1);


# // ensure to_quads returns something given colinear points

# def test_degenerate_to_quads():
#     c = CubicBez::new((0., 9.), (6., 6.), (12., 3.0), (18., 0.0));
#     quads = c.to_quads(1e-6).collect::<Vec<_>>();
#     assert_eq!(quads.len(), 1, "{:?", &quads);


def test_cubicbez_extrema():
    q = CubicBez(Point(0.0, 0.0), Point(0.0, 1.0), Point(1.0, 1.0), Point(1.0, 0.0))
    extrema = q.extrema()
    assert len(extrema) == 1
    assert abs(extrema[0] - 0.5) < 1e-6

    q = CubicBez(Point(0.4, 0.5), Point(0.0, 1.0), Point(1.0, 0.0), Point(0.5, 0.4))
    extrema = q.extrema()
    assert len(extrema) == 4


# def test_cubicbez_toquads():
#     // y = x^3
#     c = CubicBez::new((0.0, 0.0), (1.0 / 3.0, 0.0), (2.0 / 3.0, 0.0), (1.0, 1.0));
#     for i in 0..10 {
#         accuracy = 0.1f64.powi(i);
#         mut worst: f64 = 0.0;
#         for (_count, (t0, t1, q)) in c.to_quads(accuracy).enumerate() {
#             epsilon = 1e-12;
#             assert!((q.start() - c.eval(t0)).hypot() < epsilon);
#             assert!((q.end() - c.eval(t1)).hypot() < epsilon);
#             n = 4;
#             for j in 0..=n {
#                 t = (j as f64) * (n as f64).recip();
#                 p = q.eval(t);
#                 err = (p.y - p.x.powi(3)).abs();
#                 worst = worst.max(err);
#                 assert!(err < accuracy, "got { wanted {", err, accuracy);



from kurbopy import Point, Vec2, TranslateScale
import pytest


def test_point_basic():
    p = Point(1.0, 2.0)
    assert p.x == 1.0
    p.y = 3.0
    assert p.y == 3.0
    assert str(p) == "<Point x=1.0 y=3.0>"


def test_point_magic():
    p = Point(1.0, 2.0) + Vec2(1.0, 1.0)
    assert p.x == 2.0
    p = Point(1.0, 2.0) + (1.0, 1.0)
    assert p.x == 2.0


def test_point_lerp():
    p1 = Point(0.0, 0.0)
    p2 = Point(1.0, 2.0)
    mid = p1.lerp(p2, 0.5)
    assert mid.x == 0.5
    assert mid.y == 1.0


def test_point_round():
    p = Point(1.2, 2.0)
    assert p.round().x == 1.0


def test_vec2_protocol():
    p1 = Vec2(1.2, 2.0)
    p2 = Vec2(3.0, 4.0)
    assert (p1 + p2).x == 4.2
    p1 += p2
    assert p1.x == 4.2
    p1 *= 2
    assert p1.x == 8.4
    with pytest.raises(TypeError):
        p1 *= p2


def test_ts():
    ts1 = TranslateScale(Vec2(3.0, 2.0), 1.0)
    ts2 = TranslateScale(Vec2(3.0, 2.0), 1.0)
    p1 = Point(10.0, 10.0)
    moved = ts1 * p1
    assert moved.x == 13.0
    assert moved.y == 12.0


def test_common():
    from kurbopy.common import solve_cubic

    assert solve_cubic(0.1, 0.1, 0.5, 0.2) == [-2.378160678793357]

    from kurbopy.common import solve_itp

    f = lambda x: x**3 - x - 2.0
    x = solve_itp(f, 1.0, 2.0, 1e-12, 0, 0.2, f(1.0), f(2.0))
    assert abs(f(x)) < 6e-12
